

name = 'omnifig'
long_name = 'omni-fig'

version = '0.5'

url = 'https://github.com/felixludos/omni-fig'


description = 'Universal configuration system for common execution environments'

author = 'Felix Leeb'
author_email = 'felixludos.info@gmail.com'

license = 'GPL3'

readme = 'README.rst'

packages = ['omnifig']


import os
try:
	with open(os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))), 'requirements.txt'), 'r') as f:
		install_requires = f.readlines()
except:
	install_requires = ['pyyaml', 'C3Linearize', 'humpack', 'omnibelt']
del os

entry_points = {'console_scripts': 'fig = omnifig.running:entry'}



import sys, os

from .errors import NoValidProjectError
from .util import get_global_setting, resolve_order, spawn_path_options, set_global_setting
from .containers import Customizable_Infomation
from .external import include_files, include_configs, get_project_type

from .projects import Project

from omnibelt import get_printer, Registry

prt = get_printer(__name__)

_info_names = {
	'omni.fig', 'info.fig', '.omni.fig', '.info.fig',
	'fig.yaml', 'fig.yml', '.fig.yaml', '.fig.yml',
	'fig.info.yaml', 'fig.info.yml', '.fig.info.yaml', '.fig.info.yml',
	'fig_info.yaml', 'fig_info.yml', '.fig_info.yaml', '.fig_info.yml',
	'project.yaml', 'project.yml', '.project.yaml', '.project.yml',
}
_info_code = '.fig.yml'

_default_project_type = 'default'


class Profile(Customizable_Infomation):
	'''
	Generally all paths that the Profile deals with should be absolute paths as the profile operates system wide
	'''
	
	def import_info(self, raw):
		'''
		Processes the data from a yaml file and saves it to ``self``
		:param raw: data from a yaml file
		:return: None
		'''
		
		self.config_paths = raw.get('config_paths', [])
		self.src_paths = raw.get('src_paths', [])
		
		self.projects = {}
		self._project_paths = {}
		for name, path in raw.get('projects', {}).items():
			fixed = self.resolve_project_path(path)
			if fixed is not None:
				self.projects[name] = fixed
				self._project_paths[fixed] = name
			else:
				prt.warning(f'Invalid project path: {path}')
		self._loaded_projects = {}
		self._current_project = None
		
		self.default_ptype = raw.get('default_ptype', 'default')
		
		self.global_settings = raw.get('global_settings', {})
		
		self.active_projects = raw.get('active_projects', [])
		self.autoload_local = raw.get('autoload_local', True)
		
		self.save_projects_on_cleanup = raw.get('save_projects_on_cleanup', False)
		
		super().import_info(raw)
		
	@staticmethod
	def is_valid_project_path(path):
		'''Check if a path points to a project directory'''
		
		valid = lambda name: name is not None and (_info_code in name or name in _info_names)
		
		if os.path.isfile(path):
			base = os.path.basename(path)
			if valid(base):
				return path
			else:
				path = os.path.dirname(path)
		
		if os.path.isdir(path):
			for name in os.listdir(path):
				if valid(name):
					return os.path.join(path, name)
			
	
	def initialize(self, loc=None):
		'''
		Steps to execute during initialization including: updating global settings, loading configs,
		running any specified source files, loading active projects, and possibly the local project.
		
		:param loc: absolute path to current working directory (default comes from :func:`os.getcwd()`)
		:return: None
		'''
		self.update_global_settings()
		self.load_config()
		self.load_src()
		self.load_active_projects()
		if self.autoload_local:
			try:
				self.load_project(os.getcwd() if loc is None else loc)
			except NoValidProjectError:
				pass
	
	def update_global_settings(self):
		'''Updates global settings with items in :attr:`self.global_settings`'''
		for item in self.global_settings.items():
			set_global_setting(*item)
	
	def load_src(self):
		'''Run python source files in :attr:`self.src_paths`'''
		include_files(*self.src_paths)
		
	def load_config(self):
		'''Register config files and directories in :attr:`self.config_paths`'''
		include_configs(*self.config_paths)
		
	def load_active_projects(self, load_related=True):
		'''
		Load active projects in :attr:`self.active_projects` in order,
		and potentially all related projects as well
		
		:param load_related: load directly related projects
		:param all_related: recursively load all related projects
		:return: None
		'''
		for proj in self.active_projects:
			self.load_project(proj, load_related=load_related, all_related=False)
		
	def get_project(self, ident, load_related=True, all_related=False):
		'''
		Gets project if already loaded, otherwise tries to find the project path,
		and then loads the missing project
		
		:param ident: project name or path
		:param load_related: also load related projects (before this one) (dont recursively load related)
		:param all_related: recursively load all related projects (trumps ``load_related``)
		:return: project object
		'''
		
		if ident in self._loaded_projects:
			return self._loaded_projects[ident]
		
		path = self.resolve_project_path(ident)
		if path not in self._loaded_projects:
			self.load_project(path, load_related=load_related, all_related=all_related)
		return self._loaded_projects[path]
		
	def get_project_type(self, name):
		'''Gets the project type registered with the given name'''
		return get_project_type(name)
	
	def load_project(self, ident, load_related=True, all_related=False):
		'''

		:param ident: path or name of project (if name, then it must be profile.projects)
		:param load_related: also load related projects (before this one) (dont recursively load related)
		:param all_related: recursively load all related projects (trumps ``load_related``)
		:return: loaded project (or raises ``NoValidProjectError`` if none is found)
		'''
		
		path = self.resolve_project_path(ident)
		if path in self._loaded_projects:
			return self._loaded_projects[path]
		elif path is None:
			raise NoValidProjectError(ident)
		
		root = os.path.dirname(path)
		
		default_ptype = self.default_ptype
		default = self.get_project_type(default_ptype)
		
		info = default.load_raw_info(path)
		
		ptype = default.check_project_type(info)
		if ptype is None:
			ptype = default_ptype
		else:
			ptype, src_file = ptype
			if src_file is not None:
				include_files(src_file, os.path.join(root, src_file))
		
		proj_cls = self.get_project_type(ptype)
		
		project = proj_cls(raw=info)
		self._loaded_projects[path] = project
		
		if load_related or all_related:
			for related in project.get_related():
				self.load_project(related, load_related=all_related, all_related=all_related)
		
		self.set_current_project(project)
		project.initialize()
		
		return project
		
	def resolve_project_path(self, ident):
		'''
		Map/complete/fix given identifier to the project path
		
		:param ident: name or path to project
		:return: correct project path or None
		'''
		
		if not isinstance(ident, str):
			ident = ident.get_info_path()
		
		if ident in self.projects:
			return self.projects[ident]
		if ident in self._project_paths:
			return ident
		
		return self.is_valid_project_path(ident)
	
	def clear_loaded_projects(self):
		'''
		Clear all loaded projects from memory
		(this does not affect registered components or configs)
		'''
		self._loaded_projects.clear()
	
	def contains_project(self, ident): # loaded project
		return ident in self._loaded_projects or self.resolve_project_path(ident) in self._loaded_projects
	
	def track_project_info(self, name, path):
		'''Add project info to projects table :attr:`self.projects`'''
		if name in self.projects:
			prt.warning(f'Projects already contains {name}, now overwriting')
		else:
			prt.debug(f'Registering {name} in profile projects')
		self.projects[name] = path
		self._project_paths[path] = name
		self._updated = True
	
	def track_project(self, project):
		'''Add project to projects table :attr:`self.projects`'''
		return self.track_project_info(project.get_name(), project.get_info_path())
	
	def is_tracked(self, project):
		'''Check if a project is contained in projects table :attr:`self.projects`'''
		return project.get_name() in self.projects
	
	def include_project(self, project, track=True):
		'''
		Include a project instance in loaded projects table managed by this project
		and optionally track a project persistently.
		'''
		if track:
			self.track_project(project)
			
		self._loaded_projects[project.get_info_path()] = project
	
	def add_active_project(self, project):
		'''Add a project to the list of active projects'''
		name = project.get_name()
		self.active_projects.append(name)
		prt.info(f'Added active project set to {name}')
		self._updated = True
	
	def get_active_projects(self):
		'''Get a list of all projects specified as "active"'''
		return self.active_projects.copy()
	
	def set_current_project(self, project):
		'''Set the current project'''
		if isinstance(project, str):
			project = self.get_project(project)
		
		self._current_project = project
		
		prt.info(f'Current project set to {None if project is None else project.get_name()}')
	
	def get_current_project(self):
		'''Get the current project (usually loaded last and local),'''
		return self._current_project

	def cleanup(self):
		'''
		Saves project data if some has changed, and possibly also
		updates the project files of all loaded projects if they have changed.
		'''
		if self.save_projects_on_cleanup:
			for project in self._loaded_projects.values():
				project.cleanup()
		super().cleanup()
	




import sys, os
import yaml
from collections import namedtuple, OrderedDict

from omnibelt import get_printer, get_now

prt = get_printer(__name__)


class WrongInfoContainerType(Exception):
	def __init__(self, mtype, mtype_src=None):
		super().__init__(f'Reload manager using custom subclass: {mtype}')
		self.mtype = mtype
		self.mtype_src = mtype_src
	
	def get_mtype(self):
		return self.mtype
	
	def get_mtype_src(self):
		return self.mtype_src

class Customizable_Infomation:
	'''
	Keeps track of some data extracted from a yaml.
	Consequently a manager is entirely defined by a path to the yaml file
	which holds the source data, however managers can be subclassed to define
	custom behavior in how that data is used.
	
	Main uses include projects and profiles.
	
	Generally any attribute holds data that can be exported,
	however attributes that start with "_" are skipped
	
	'''
	register = None
	mtype = None
	
	required_attrs = []
	recommended_attrs = []
	
	def __init_subclass__(cls, mtype=None):
		cls.mtype = mtype
		if cls.register is not None:
			cls.register.new(mtype, cls)
	
	# @classmethod
	# def get_mtype(cls):
	# 	return cls.mtype
	
	def __init__(self, name=None, raw=None, path=None):
		self._updated = False
		
		if raw is None and path is not None:
			raw = self.load_raw_info(path)
		
		if raw is None:
			raw = {}
		
		self.import_info(raw)
			
		if name is not None:
		# 	prt.warning(f'No name provided for the profile')
		# else:
			self.name = name
	
	@staticmethod
	def find_manager_type(raw):
		'''
		Based on the raw info, confirm that this is the right type to use as the manager.
		This enables using custom subclasses of Manager
		
		:param raw: src info (usually from some yaml file)
		:return: must throw a WrongManager exception if a different one should be used
		'''
		
		mtype = raw.get('type', None)
		
		if mtype is not None:
			mtype_src = raw.get('type_src', None)
			raise WrongInfoContainerType(mtype, mtype_src)
		
	@staticmethod
	def load_raw_info(path):
		with open(path, 'r') as f:
			raw = yaml.safe_load(f)
		raw['info_path'] = path # automatically set info_path to the path
		raw['info_dir'] = os.path.dirname(path)
		return raw
	
	def get_name(self):
		return self.name
	
	def __str__(self):
		return f'{self.get_name()}'
	
	def __repr__(self):
		return f'{self.get_mtype()}:{self.get_name()}'
	
	def get_info_path(self):
		return getattr(self, 'info_path', None)
	
	def import_info(self, raw={}):
		self.find_manager_type(raw)
		
		for key in raw:
			if key not in self.__dict__:
				setattr(self, key, raw[key])
		
		for key in self.required_attrs:
			if getattr(self, key) is None:
				prt.error(f'{key} not found in {self.__class__.__name__} {self}')

		for key in self.recommended_attrs:
			if getattr(self, key) is None:
				prt.warning(f'{key} not found in {self.__class__.__name__} {self}')
	
	def export_info(self, path=None):
		'''
		Saves info to yaml (by default where it was loaded from)

		Filters out any entries with keys that start with '_' or have no value (None)
		
		:return: path to export if one is provided, otherwise all the data that would have been exported as a dict
		'''
		
		if path is None:
			path = self.get_info_path()
		
		data = {}
		for k, v in self.__dict__.items():
			if v is not None and k[0] != '_':
				data[k] = v
		
		if self._updated: # something was changed
			data['last_update'] = get_now()
		
		if path is None:
			prt.warning('No export path found, so there is nowhere to save the info')
			return
		
		with open(path, 'w') as f:
			yaml.dump(data, f)
		
		prt.debug(f'Exported {repr(self)} to {path}')
		return path
	
	def cleanup(self):
		if self._updated:
			self.export_info()
	



import sys
import traceback

from omnibelt import get_printer

from .errors import ConfigNotFoundError
from .config import get_config
from .modes import Run_Mode
from .rules import Meta_Rule
from .registry import Component

DEBUG_CODE = 'd'
DEBUG_NAME = 'debug'
DEBUG_MODE_NAME = f'run_mode/{DEBUG_NAME}'

prt = get_printer(__name__)

@Meta_Rule(DEBUG_NAME, priority=100, code=DEBUG_CODE, description='Switch to debug mode')
def debug_rule(meta, config):
	'''
	When activated, this rule changes the run mode to ``run_mode/debug`` and updates the config to include
	
	:param meta: meta config object
	:param config: full config object
	:return: config object updated with debug config (if available)
	'''
	
	debug = meta.pull(DEBUG_NAME, False, silent=True)
	if debug:
		meta.push('_type', DEBUG_MODE_NAME, silent=True)
		print('Switched to debug run mode')
	
		try:
			debug = get_config(DEBUG_NAME)
		except ConfigNotFoundError:
			if not meta.pull('silent', silent=True):
				prt.warning('No config "debug" was found')
		else:
			config.update(debug)
		
		meta.push('debug', False, silent=True) # to prevent running this rule multiple times for the same config

	return config


@Component(DEBUG_MODE_NAME)
class Debug_Mode(Run_Mode):
	'''
	Behaves just like the default run mode, excapt if the script raises an exception,
	either the ``ipdb`` :func:`post_mortem` debugger is activated, or if this is already
	running in a debugger (checked using :code:`sys.gettrace() is not None`) then
	``ipdb`` is not activated.
	'''

	def run(self, meta, config):
		'''Calls ``ipdb`` to activate debugger if an exception is raised when running the script.'''
		
		try:
			return super().run(meta, config)
		except KeyboardInterrupt:
			extype, value, tb = sys.exc_info()
			traceback.print_exc()

		except Exception as e:
			if sys.gettrace() is None:
				import ipdb
				extype, value, tb = sys.exc_info()
				traceback.print_exc()
				ipdb.post_mortem(tb)
			else:
				print('[Skipping debug]')
				raise e
		
	






from .errors import ScriptNotFoundError
from .registry import Component, get_script
from .util import autofill_args

from omnibelt import get_printer

prt = get_printer(__name__)


@Component('run_mode/default')
class Run_Mode:
	'''
	Run modes are used to actually execute the script specified by the user in the run command.
	
	It is recommended to register all run_modes with a ``run_mode/`` prefix, but not required.
	'''
	def __init__(self, A):
		self.silent = A.pull('silent', True, silent=True)
	
	@staticmethod
	def get_script_info(script_name):
		'''Given the name of the registered script, this returns the corresponding entry in the registry'''
		return get_script(script_name)
	
	
	def run(self, meta, config):
		'''
		When called this should actually execute the registered script whose name is specified under meta.script_name.

		:param meta: meta config - contains script_name
		:param config: config for script
		:return: output of script
		'''
		
		script_name = meta.pull('script_name', silent=self.silent)
		script_info = self.get_script_info(script_name)
		
		if script_info is None:
			raise ScriptNotFoundError(script_name)
		
		script_fn = script_info.fn
		
		if script_info.use_config:
			return script_fn(config)
		return autofill_args(script_fn, config)




from .registry import AutoScript, Script, Component, AutoComponent, Modifier, AutoModifier, Modification, \
	create_component, register_component, register_modifier, \
	view_component_registry, view_modifier_registry, view_script_registry
from .config import get_config, MissingConfigError
from .external import register_config, register_config_dir
from .running import entry, main, run, quick_run, initialize, cleanup
from .loading import get_project, get_profile
from .modes import Run_Mode
from .rules import Meta_Rule
from .debug import Debug_Mode
from .help import help_message

from humpack import AbortTransaction

#

import os
__info__ = {'__file__':os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py')}
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py'), 'r') as f:
	exec(f.read(), __info__)
del os
del __info__['__file__']
__author__ = __info__['author']
__version__ = __info__['version']




import sys, os
import inspect
import time

from omnibelt import primitives, get_printer, \
	get_global_settings, set_global_setting, get_global_setting, get_now, resolve_order, \
	spawn_path_options
from omnibelt.logging import global_settings

LIB_PATH = os.path.dirname(__file__)

global_settings.update({
	
	# naming settings
	'default_entry_name': 'default',
	'profile_env_var': 'FIG_PROFILE',
	'profile_src_env_var': 'FIG_PROFILE_SRC',
	'init_env_var': 'FIG_INIT',
	
	# default infos
	'default_profile': None,
	
	'allow_profile_sources': True,
	'use_last_entry_as_default': True,
	'default_ptype': 'default',
	
	# 'resolution_order': [ # for scripts, configs, components, and modifiers
	# 	 'local', # check the current working dir if its in a project
	#      'active-project', # currently active project
	#      'related-projects', # any projects related to the active project
	#      'init' # scripts loaded through init (no project)
	#      ],
})

prt = get_printer(__name__)


def autofill_args(fn, config, aliases=None, run=True):

	params = inspect.signature(fn).parameters

	args = []
	kwargs = {}

	for n, p in params.items():

		order = [n]
		if aliases is not None and n in aliases: # include aliases
			order.extend('<>{}'.format(a) for a in aliases[n])
		if p.default != inspect._empty:
			order.append(p.default)
		elif p.kind == p.VAR_POSITIONAL:
			order.append(())
		elif p.kind == p.VAR_KEYWORD:
			order.append({})

		arg = config.pull(*order)

		if p.kind == p.POSITIONAL_ONLY:
			args.append(arg)
		elif p.kind == p.VAR_POSITIONAL:
			args.extend(arg)
		elif p.kind == p.VAR_KEYWORD:
			kwargs.update(arg)
		else:
			kwargs[n] = arg
	if run:
		return fn(*args, **kwargs)
	return args, kwargs










class NoValidProjectError(Exception):
	'''Raised when no project is found for the given identifier (which should be the name or path to the project)'''
	def __init__(self, ident):
		super().__init__(f'Not a valid project: {ident}')

class YamlifyError(Exception):
	'''Raised when an object is unable to be turned into a yaml object (primitives, dicts, lists)'''
	def __init__(self, obj):
		super().__init__('Unable to yamlify: {} (type={})'.format(obj, type(obj)))
		self.obj = obj

class UnknownActionError(Exception):
	'''Raised when trying to record an unrecognized action with the config object'''
	pass
	
class ScriptNotFoundError(Exception):
	'''Raised when trying to run a script that has not been registered'''
	pass
	
class ConfigNotFoundError(Exception):
	'''Raised when a config parameter is not found and no viable defaults are provided'''
	def __init__(self, ident):
		super().__init__(f'Unknown config: {ident}')

class MissingConfigError(Exception):
	'''Raised when a config parameter was not found, and no viable defaults were provided'''
	def __init__(self, key):
		super().__init__(key)

class InvalidKeyError(Exception):
	'''Only raised when a key cannot be converted to an index for a :class:`ConfigList`'''
	pass

# class NoConfigFound(Exception):
# 	def __init__(self):
# 		super().__init__(
# 			'Either provide a config name/path as the first argument, or set your $FOUNDATION_CONFIG environment variable to a config name/path')






import sys, os

from .external import include_files
from .rules import meta_rule_fns, view_meta_rules
from .config import get_config, process_raw_argv
from .loading import get_profile

from omnibelt import get_printer, resolve_order

prt = get_printer(__name__)

PRINCEPS_NAME = 'FIG_PRINCEPS_PATH'
_disable_princeps = False


def entry(script_name=None):
	'''
	Recommended entry point when running a script from the terminal.
	This is also the entry point for the ``fig`` command.
	
	This collects the command line arguments in ``sys.argv`` and overrides the
	given script with ``script_name`` if it is provided
	
	:param script_name: script to be run (may be set with arguments) (overrides other arguments if provided)
	:return: the output of the script that is to be run
	'''
	argv = sys.argv[1:]
	return main(*argv, script_name=script_name)

def main(*argv, script_name=None):
	'''
	Runs the desired script using the provided ``argv`` which are treated as command line arguments
	
	Before running the script, this function initializes ``omni-fig`` using :func:`initialize()`,
	and then cleans up after running using :func:`cleanup()`.
	
	:param argv: raw arguments as if passed in through the terminal
	:param script_name: name of registered script to be run (may be set with arguments) (overrides other arguments if provided)
	:return: output of script that is run
	'''
	initialize()
	config = process_argv(argv, script_name=script_name)
	
	out = run(config=config)
	
	cleanup()
	
	return out



def process_argv(argv=(), script_name=None):
	'''
	Parses the command line arguments to identify the meta arguments, script name
	(optionally overridden using ``script_name``), and config args
	
	From that, this builds the config and meta config object.
	
	:param argv: list of all command line arguments to parse in order
	:param script_name: optional script name to override any script specified in ``argv``
	:return: config object (containing meta config under ``_meta``)
	'''
	
	# check for meta args and script name
	
	meta = {}
	
	waiting_key = None
	waiting_meta = 0
	
	remaining = []
	for i, arg in enumerate(argv):
		
		if waiting_meta > 0:
			if waiting_key in meta and isinstance(meta[waiting_key], list):
				meta[waiting_key].append(process_raw_argv(arg))
			else:
				meta[waiting_key] = process_raw_argv(arg)
			waiting_meta -=1
			if waiting_meta == 0:
				waiting_key = None
	
		elif '-' == arg[0] and not arg.startswith('--'):
			text = arg[1:]
			for rule in view_meta_rules():
				name = rule.name
				code = rule.code
				if code is not None and text.startswith(code):
					text = text[len(code):]
					num = rule.num_args
					if num:
						if len(text):
							raise Exception(f'Can\'t combine multiple meta-rules if they require params: {code} in {text}')
						waiting_key = name
						waiting_meta = num
						if num > 1:
							meta[waiting_key] = []
					else:
						meta[name] = True
				if not len(text):
					break
					
		elif arg == '_' or script_name is not None:
			remaining = argv[i:]
			break
			
		else:
			script_name = arg
	
	if script_name is not None:
		meta['script_name'] = script_name
	
	# call get_config
	config = get_config(*remaining)
	config.sub('_meta').update(meta)
	
	return config


def initialize(*projects, **overrides):
	'''
	Initializes omni-fig by running the "princeps" file (if one exists),
	loading the profile, and any active projects. Additionally loads the
	project in the current working directory (by default).
	
	Generally, this function should be run before running any scripts, as it should register all
	necessary scripts, components, and configs when loading a project. It is automatically called
	when running the :func:`main()` function (ie. running through the terminal). However, when
	starting scripts from other environments (such as in a jupyter notebook), this should be called
	manually after importing ``omnifig``.
	
	:param projects: additional projects that should be initialized
	:param overrides: settings to be checked before defaulting to ``os.environ`` or global settings
	:return: None
	'''
	
	# princeps script
	princeps_path = resolve_order(PRINCEPS_NAME, overrides, os.environ)
	if not _disable_princeps and princeps_path is not None:
		try:
			include_files(princeps_path)
		except Exception as e:
			prt.critical(f'Failed to run princeps: {princeps_path}')
			raise e
	
	# load profile
	profile = get_profile(**overrides)
	
	# load project/s
	profile.initialize()
	
	for proj in projects:
		profile.load_project(proj)
	

def cleanup(**overrides):
	'''
	Cleans up the projects and profile, which by default just updates the project/profile info
	yaml file if new information was added to the project/profile.
	
	Generally, this should be run after running any desired scripts.
	
	:param overrides: settings to check before defaulting to global settings or ``os.environ``
	:return: None
	'''
	get_profile(**overrides).cleanup()



def run(script_name=None, config=None, **meta_args):
	'''
	This actually runs the script given the ``config`` object.
	
	Before starting the script, all meta rules are executed in order of priority (low to high)
	as they may change the config or script behavior, then the run mode is created, which is
	then called to execute the script specified in the config object (or manually overridden
	using ``script_name``)
	
	:param script_name: registered script name to run (overrides what is specified in ``config``)
	:param config: config object (usually created with :func:`get_config()` (see :ref:`config:Config System`)
	:param meta_args: Any additional meta arguments to include before running
	:return: script output
	'''
	if config is None:
		config = get_config()
	
	if script_name is not None:
		config.push('_meta.script_name', script_name, overwrite=True, silent=True)
	for k, v in meta_args.items():
		config.push(f'_meta.{k}', v, overwrite=True, silent=True)
	# config._meta.update(meta_args)
	
	for rule in meta_rule_fns():
		config = rule(config.sub('_meta'), config)
	
	config.push('_meta._type', 'run_mode/default', overwrite=False, silent=True)
	silent = config.pull('_meta._quiet_run_mode', True, silent=True)
	mode = config.pull('_meta', silent=silent)
	# config = mode.process(config)
	
	return mode.run(config.sub('_meta'), config)


def quick_run(script_name, *parents, **args):
	'''
	Convenience function to run a simple script without a given config object,
	instead the config is entirely created using the provided ``parents`` and ``args``.
	
	:param script_name: name of registered script that is to be run
	:param parents: any names of registered configs to load
	:param args: any additional arguments to be provided manually
	:return: script output
	'''
	config = get_config(*parents, **args)
	return run(script_name, config)








import sys, os
import importlib
import importlib.util

from omnibelt import get_printer, Registry, Entry_Registry

from .errors import ConfigNotFoundError

prt = get_printer(__name__)

# region Source files

_loaded_files = {}
_load_counter = 0
def include_files(*paths):
	'''
	Executes all provided paths to python files that have not already been run.
	
	:param paths: paths to python files to be executed (if not already)
	:return: None
	'''
	global _load_counter
	
	for path in paths:
		if os.path.isfile(path):
			apath = os.path.abspath(path)
			if apath not in _loaded_files:
				prt.debug(f'Loading {apath}')
				spec = importlib.util.spec_from_file_location(f"load{_load_counter}", apath)
				mod = importlib.util.module_from_spec(spec)
				spec.loader.exec_module(mod)
				# mod.MyClass()
				_loaded_files[apath] = mod
				_load_counter += 1

def include_package(*packages):
	'''
	Imports packages based on their names
	
	:param packages: list of package names to be imported
	:return: None
	'''
	for pkg in packages:
		importlib.import_module(pkg)

# endregion


# region config

class _Config_Registry(Entry_Registry, components=['path', 'project']):
	pass
_config_registry = _Config_Registry()


def register_config(name, path, project=None):
	'''
	Registers a single config yaml file
	
	:param name: name to associate with this config
	:param path: absolute path to config
	:param project: project associated with this config
	:return: None
	'''
	assert os.path.isfile(path), 'Cant find config file: {}'.format(path)
	
	prt.debug(f'Registering config {name}')
	if name in _config_registry:
		prt.warning(f'A config with name {name} has already been registered, now overwriting')
	
	_config_registry.new(name, path=path, project=project)
	
	if project is not None:
		project.new_config(name)


def register_config_dir(path, recursive=False, prefix=None, joiner='/', project=None):
	'''
	Registers all yaml files found in the given directory (possibly recursively)
	
	When recusively checking all directories inside, the internal folder hierarchy is preserved
	in the name of the config registered, so for example if the given ``path`` points to a
	directory that contains a directory ``a`` and two files ``f1.yaml`` and ``f2.yaml``:
	
	Contents of ``path`` and corresponding registered names:
	
		- ``f1.yaml`` => ``f1``
		- ``f2.yaml`` => ``f2``
		- ``a/f3.yaml`` => ``a/f3``
		- ``a/b/f4.yaml`` => ``a/b/f3``
	
	If a ``prefix`` is provided, it is appended to the beginning of the registered names
	
	:param path: path to root directory to search through
	:param recursive: search recursively through sub-directories for more config yaml files
	:param prefix: prefix for names of configs found herein
	:param joiner: string to merge directories when recursively searching (default ``/``)
	:param project: project object to associate with all configs registered here
	:return: None
	'''
	assert os.path.isdir(path)
	for fname in os.listdir(path):
		parts = fname.split('.')
		candidate = os.path.join(path, fname)
		if os.path.isfile(candidate) and len(parts) > 1 and parts[-1] in {'yml', 'yaml'}:
			name = parts[0]
			if prefix is not None:
				name = joiner.join([prefix, name])
			register_config(name, os.path.join(path, fname), project=project)
		elif recursive and os.path.isdir(candidate):
			newprefix = fname if prefix is None else joiner.join([prefix, fname])
			register_config_dir(candidate, recursive=recursive, prefix=newprefix, joiner=joiner, project=project)

def include_configs(*paths, project=None):
	'''
	Convenience function which can be given paths to directories or individual files,
	all of which will be registered (recursively)
	
	:param paths: paths to yaml config files or directories to be registered
	:param project: project to associate with all configs registered here
	:return: None
	'''
	for path in paths:
		if os.path.isdir(path):
			register_config_dir(path, recursive=True, project=project)
		elif os.path.isfile(path):
			fname = os.path.basename(path)
			parts = fname.split('.')
			if len(parts) > 1 and parts[-1] in {'yml', 'yaml'}:
				register_config(parts[0], path, project=project)

def view_config_registry(): # TODO: clean this up - return view not copy of the full registry
	'''Return a copy of the full config registry'''
	return _config_registry.copy()


def find_config_path(name):
	'''
	Given a name or path, find the associated config file
	by checking the registry in possible
	
	:param name: path or name of config to find
	:return: full absolute path of config (raising :class:`ConfigNotFoundError` if nothing is found)
	'''
	if os.path.isfile(name):
		return name

	reg = _config_registry

	if name in reg:
		return reg[name].path
	elif os.path.isfile(name):
		return name
	raise ConfigNotFoundError(name)


# endregion

# region project types

_ptype_registry = Registry()


def register_project_type(name, cls):
	'''
	Project types allow users to customize the behavior of project objects

	:param name: identifier of this project type
	:param cls: project type class
	:return: None
	'''

	_ptype_registry.new(name, cls)


def get_project_type(name):
	'''Gets the project type associated with that name, otherwise returns None'''
	if name not in _ptype_registry:
		prt.error(f'Project type: {name} not found')
	return _ptype_registry.get(name, None)


def view_project_types():
	'''Returns a copy of the full project type registry'''
	return _ptype_registry.copy()

# endregion



from heapq import heapify, heappop

from omnibelt import get_printer, Entry_Registry

prt = get_printer(__name__)

class _Rules_Registry(Entry_Registry, components=['fn', 'priority', 'code', 'num_args', 'description']):
	pass
_rules_registry = _Rules_Registry()


def Meta_Rule(name: str, priority: int = 0, code: str = None,
              expects_args: int = 0, description: str = None):
	'''
	Decorator used to register meta rules.

	:param name: Name of the meta rule (and key in meta config for any required arguments)
	:param priority: priority of this meta rule (the higher it is the earlier it is run relative to other meta rules)
	:param code: Command line code (usually a single letter) preceded by "-" to activate this meta rule
	:param expects_args: number of command line arguments required when activating
	:param description: short description of what this rule does for the help message
	:return: decorator function to register the meta rule
	'''
	
	def _reg(fn):
		register_meta_rule(name, fn=fn, priority=priority, code=code, expects_args=expects_args,
		                   description=description)
		return fn
	
	return _reg


def register_meta_rule(name, fn, priority=0, code=None, expects_args=0, description=None):
	'''
	Meta Rules are expected to be a callable of the form:
	
	config <= fn(meta, config)
	
	where ``meta`` is the meta config, and the output should be the processed ``config``
	
	:param name: name of the new rule
	:param fn: callable to enforce rule
	:param priority: priority of this rule (rules are enforced by priority high to low)
	:param code: (optional) single letter code used to activate rule from the terminal
	:param expects_args: number of arguments expected by this meta rule through the terminal
	:param description: one line description of what the rule does
	:return: None
	'''
	
	_rules_registry.new(name, fn=fn, priority=priority, code=code, num_args=int(expects_args),
	                    description=description)

def view_meta_rules():
	'''Returns an iterator over all registered meta rule entries in order of priority (low to high)'''
	rules = [(-r.priority, r) for r in _rules_registry.values()]
	heapify(rules)
	
	while len(rules):
		yield heappop(rules)[1]

def meta_rule_fns():
	'''Returns an iterator over all meta rule functions in order of priority'''
	for rule in view_meta_rules():
		yield rule.fn





import sys, os
import yaml
from argparse import Namespace

from .containers import Customizable_Infomation
from .external import include_files, register_project_type, include_configs, include_package

from omnibelt import get_printer, get_now

prt = get_printer(__name__)

class Project(Customizable_Infomation):
	'''
	Projects are used to group code into packets with specific config files that should be
	loaded all together. A project must contain a yaml file named ``.fig.yml`` in the
	root directory of the project (aka the "project directory"), and all paths in
	that yaml file should be relative to the project directory.
	
	Generally there are two kinds of projects: "packages" and "loose" projects.
	Package projects are meant to be installed and used as a library, while loose
	projects may just be a series of python files.
	
	This class may also be subclassed to change the behavior of projects (such as changing the loading),
	in fact, any subclasses of this class can automatically be registered when providing a name for the
	project type in the class definition.
	'''

	# required_attrs = ['name', 'author', 'info_path']
	# recommended_attrs = ['url', 'version', 'license', 'description']
	
	def __init_subclass__(cls, ptype=None):
		'''Subclasses can automatically be registered if a ``ptype`` for the registry is provided'''
		cls.ptype = ptype
		register_project_type(ptype, cls)
	
	def __str__(self):
		return self.get_name()
	
	def __repr__(self):
		return f'Project({self.get_name()})'
	
	def get_name(self):
		'''Gets the project name'''
		return self.name
	def get_info_path(self):
		'''Gets the project directory'''
		return self.info_path
	
	@staticmethod
	def check_project_type(raw):
		'''
		Based on raw project info, check if this project expects a certain project type, if so
		optionally provide a path to the source file of the project type.

		:param raw: raw project info
		:return: None or tuple with project type identifier and path to source file (or None)
		'''
		ptype = raw.get('project_type', None)
		if ptype is not None:
			return ptype, raw.get('ptype_src_file', None)
	
	def import_info(self, raw):
		'''
		Given the raw info loaded from a yaml file, this function checks integrates the information into
		the project object (self).
		
		:param raw: dictionary of info (usually loaded from a yaml)
		'''
		
		
		# region path
		
		self.info_path = raw['info_path']
		self.root = os.path.dirname(self.info_path)
		
		# endregion
		
		if 'py_info' in raw:
			info = {'__file__': os.path.join(self.root, raw['py_info'])}
			with open(info['__file__'], 'r') as f:
				exec(f.read(), info)
			del info['__file__']
			raw.update(info)
		
		# region info
		
		self.name = raw.get('name', None)
		
		self.author = raw.get('author', None)
		self.authors = raw.get('authors', None)
		if self.authors is None and self.author is not None:
			self.authors = [self.author]
		
		self.github = raw.get('github', None)
		self.url = raw.get('url', None)
		if self.github is not None and self.url is None:
			self.url = f'https://github.com/{self.github}'
		
		self.version = raw.get('version', None)
		self.license = raw.get('license', None)
		self.use = raw.get('use', 'leaf') # {leaf, package}
		
		self.add_to_path = raw.get('add_to_path', True)
		
		self.description = raw.get('description', None)
		
		for key in raw:
			if key not in self.__dict__:
				prt.info(f'Found optional project info: {key}')
				setattr(self, key, raw[key])
		
		# endregion
		# region related
		
		self.related = raw.get('related', []) # should be a list of idents (paths or names in profile)
		# self.dependenies = raw.get('dependency', []) # names of projects that must be loaded before this one
		
		# endregion
		# region components
		
		self.last_update = raw.get('last_update', None)
		self.conda_env = raw.get('conda', None) # TODO
		
		self.config_paths = raw.get('configs', [])
		src = raw.get('src', None)
		self.src_paths = raw.get('src_paths', [])
		if src is not None:
			self.src_paths = self.src_paths + [src]
		self.src_packages = raw.get('src_packages', [])
		
		self.scripts = raw.get('scripts', []) # TODO: maybe track contents
		self.configs = raw.get('configs', [])
		self.components = raw.get('components', [])
		self.modifiers = raw.get('modifiers', [])
		
		self.no_auto_config = raw.get('no_auto_config', None)
		if self.no_auto_config is None or not self.no_auto_config:
			contents = set(os.listdir(self.root))
			auto_names = {'config', 'configs'}
			for aname in auto_names:
				if aname in contents:
					auto_config = os.path.join(self.root, aname)
					if os.path.isdir(auto_config) and auto_config not in self.config_paths:
						self.config_paths.append(auto_config)
			
		# endregion
		
		prt.debug(f'Finished loading project info for {self.name} ({self.info_path})')
		
		super().import_info(raw)
		
	def initialize(self):
		'''
		This loads the project, primarily by registering any specified config files,
		importing specified packages, and finally running any provided source files
		
		:return: None
		'''
		if self.add_to_path:
			sys.path.append(self.root)
		self.load_configs()
		self.load_src()
		
	def load_configs(self):
		'''Registers all specified config files and directories'''
		if len(self.config_paths):
			include_configs(*self.config_paths, project=self)
			
	def load_src(self):
		'''Imports all specified packages and runs the specified python files'''
		include_package(*self.src_packages)
		include_files(*[os.path.join(self.root, src) for src in self.src_paths])
	
	def get_related(self):
		'''Returns a list of project names of all projects that should be loaded prior to this one'''
		return self.related
		
	def export_info(self, path=None):
		'''
		Saves info to yaml (by default where it was loaded from)
		
		Filters out any entries with keys that start with '_' or have no value (None)
		'''
		
		data = {}
		for k,v in self.__dict__.items():
			if v is not None and k[0] != '_':
				data[k] = v
				
		if self._updated:
			data['last_update'] = get_now()
		
		if path is None:
			assert 'info_path' in data and data['info_path'] is not None, 'No path for export found'
			path = data['info_path']
		data['info_path'] = path
		
		with open(path, 'w') as f:
			yaml.dump(data, f)
	
		prt.debug(f'Finished exporting project info of {self.name} to {self.info_path}')

	def new_script(self, name):
		'''Adds the specified script name to this project for record keeping'''
		if name not in self.scripts:
			self.scripts.append(name)
		self._updated = True

	def new_config(self, name):
		'''Adds the specified config file to this project for record keeping'''
		if name not in self.configs:
			self.configs.append(name)
		self._updated = True
	
	def new_component(self, name):
		'''Adds the specified component name to this project for record keeping'''
		if name not in self.components:
			self.components.append(name)
		self._updated = True
	
	def new_modifier(self, name):
		'''Adds the specified modifier name to this project for record keeping'''
		if name not in self.modifiers:
			self.modifiers.append(name)
		self._updated = True

register_project_type('default', Project)




import sys, os
from tabulate import tabulate

from .external import view_config_registry
from .registry import get_script, view_script_registry
from .rules import Meta_Rule, view_meta_rules


_default_help_msg = '''
Usage: fig [-<meta>] {script} [<configs>...] [--<args>]

{scripts}

+------+
| meta |
+------+
Optional meta arguments (usually specified with a single letter)
{metas}

+---------+
| configs |
+---------+
Specify any registered configs to merge

{configs}

+------+
| args |
+------+
Any additional arguments specified manually
as key-value pairs (keys with "--")
'''

_default_script_info = '''+--------+
| script |
+--------+
Specify a registered script name to run
{available}'''

_script_sel = '''Script: {script}
{doc}'''


@Meta_Rule('help', priority=99, code='h', description='Display this help message')
def help_message(meta, config):
	'''
	When activated, this rule prints out help message for the ``fig`` command, which includes
	a list of all registered scripts, meta rules, and configs that have been loaded.
	
	:param meta: meta config object
	:param config: config object
	:return: [system exit, with code 0]
	'''
	
	show_help = meta.pull('help', False, silent=True)
	
	if not show_help:
		return config
	
	name = meta.pull('script_name', None, silent=True)
	num = meta.pull('show_args', 4, silent=True)
	
	metas = [(f'-{r.code}', r.name, '-' if r.description is None else r.description)
	         for r in view_meta_rules() if r.code is not None][:num]
	
	minfo = tabulate(metas, headers=['Code', 'Name', 'Description'], ) # tablefmt="plain"
	
	scripts = [(s.name, '-' if s.description is None else s.description)
	           for s in view_script_registry().values()]#[:num]
	
	configs = [c.name for c in view_config_registry().values()]
	
	end = ''
	if len(configs) > num:
		end = f' ... [{len(configs)} items]'
	
	cinfo = ', '.join(configs)
	cinfo = f'Registered configs: {cinfo}{end}'
	
	if name is None:
		name = '<script>'
		
		if len(scripts) == 0:
			sinfo = '\n  - No scripts registered -'
		else:
			sinfo = tabulate(scripts, headers=['Name', 'Description'],) #tablefmt="plain"
		
		sinfo = _default_script_info.format(available=sinfo)
	
	else:
		
		info = get_script(name)
		doc = info.fn.__doc__
	
		if doc is None or len(doc) == 0:
			doc = '[no docstring]\n'
	
		sinfo = _script_sel.format(script=name, doc=doc)
		
	print(_default_help_msg.format(script=name, scripts=sinfo, metas=minfo, configs=cinfo))
	
	sys.exit(0)





import sys, os

from .profiles import Profile

from omnibelt import get_printer, resolve_order

prt = get_printer(__name__)


PROFILE_NAME = 'FIG_PROFILE'


_profile_cls = Profile

def set_profile_cls(cls):
	'''Set the class used when loading the profile'''
	global _profile_cls
	prt.debug(f'Profile cls used: {_profile_cls.__name__}')
	_profile_cls = cls

def get_profile_cls():
	'''Get the class used when loading the profile'''
	return _profile_cls


_profile = None
def set_profile(profile):
	'''Set the loaded profile object'''
	global _profile
	prt.debug(f'Profile set to: {profile}')
	_profile = profile


def load_profile(**overrides):
	'''
	Load the profile with the yaml file with the path found
	with the environment variable ``FIG_PROFILE``
	
	:param overrides: Any additional overrides to use instead of checking the environment variables
	:return: loaded profile object
	'''
	global _profile
	
	profile_path = resolve_order(PROFILE_NAME, overrides, os.environ)
	if profile_path is None:
		prt.warning(f'No profile path provided (can be set with environment variable: {PROFILE_NAME})')
	
	_profile = _profile_cls(path=profile_path)
	
	if profile_path is None:
		prt.info('No profile path found')
	
	return _profile


def get_profile(**overrides):
	'''Returns current profile (which gets loaded if there is None)'''
	if _profile is None:
		load_profile(**overrides)
	return _profile

def get_project(ident):
	'''Checks the profile to return (and possibly load) a project given the name or path ``ident``'''
	profile = get_profile()
	if profile is not None:
		return profile.get_project(ident)


class set_current_project:
	'''Context manager to change set the current project in the context'''
	def __init__(self, project=None):
		self.prev = get_current_project()
		self.new = project
		
		profile = get_profile()
		profile.set_active_project(project)
	
	def __enter__(self):
		pass
	
	def __exit__(self, exc_type, exc_value, exc_traceback):
		set_current_project(self.prev)


def clear_current_project():
	'''Unset the current project (setting it to None)'''
	set_current_project()


def get_current_project():
	'''Get the current project, assuming a profile is loaded, otherwise returns None'''
	profile = get_profile()
	if profile is not None:
		return profile.get_current_project()









import sys, os
import inspect
from collections import namedtuple
import importlib.util

from .errors import MissingConfigError
from .loading import get_current_project
from .util import autofill_args

from omnibelt import Registry, Entry_Registry, get_printer, get_global_setting

prt = get_printer(__name__)


# region Scripts

class _Script_Registry(Entry_Registry, components=['fn', 'use_config', 'description', 'project']):
	pass
_script_registry = _Script_Registry()

def register_script(name, fn, use_config=False, description=None):
	'''
	Function to register a script
	
	:param name: name of script
	:param fn: script function (usually a callable that expects the config object)
	:param use_config: :code:`True` if the config should be passed as only arg when calling the script function, otherise it will automatically pull all arguments in the script function signature
	:param description: a short description of what the script does
	:return:
	'''
	# prt.debug(f'Registering script {name}')
	# if name in _script_registry:
	# 	prt.warning(f'A script with name {name} has already been registered, now overwriting')
	
	project = get_current_project()
	_script_registry.new(name, fn=fn, use_config=use_config, description=description, project=project)
	
	if project is not None:
		project.new_script(name)
	
def get_script(name):
	'''Returns the entry for registered with the given name, or :code:`None` if not found'''
	return _script_registry.get(name, None)
	

def Script(name, use_config=True, description=None):
	'''
	Decorator to register a script
	
	:param name: name of script
	:param use_config: :code:`True` if the config should be passed as only arg when calling the script function, otherise it will automatically pull all arguments in the script function signature
	:param description: a short description of what the script does
	:return: decorator function expecting a callable
	'''
	def _reg_script(fn):
		nonlocal name, use_config
		register_script(name, fn, use_config=use_config, description=description)
		return fn
	
	return _reg_script

def AutoScript(name, description=None):
	'''
	Convienence decorator to register scripts that automatically extract
	relevant arguments from the config object
	
	:param name: name of the script
	:param description: a short description of what the script does
	:return: decorator function expecting a callable that does not expect the config as argument (otherwise use :func:`Script`)
	'''
	return Script(name, use_config=False, description=description)

# endregion

# region Components

class _Component_Registry(Entry_Registry, components=['create_fn', 'project']):
	pass
_component_registry = _Component_Registry()

def register_component(name, create_fn):
	'''
	create_fn takes a single input - a Config object
	The config object is guaranteed to have at least one entry with key "_type" and the value is the same as
	the registered name of the component.

	:param name: str (should be unique to this component)
	:param create_fn: callable accepting one arg (a Config object) (these should usually be classes)
	'''
	# assert name not in _reserved_names, '{} is reserved'.format(name)
	# prt.debug(f'Registering component {name}')
	# if name in _component_registry:
	# 	prt.warning(f'A component with name {name} has already been registered, now overwriting')
	
	project = get_current_project()
	_component_registry.new(name, create_fn=create_fn, project=project)

	if project is not None:
		project.new_component(name)
	

def Component(name=None):
	'''
	Decorator to register a component

	NOTE: components should usually be types/classes to allow modifications

	:param name: if not provided, will use the __name__ attribute.
	:return: decorator function
	'''
	
	def _cmp(cmp):
		nonlocal name
		if name is None:
			name = cmp.__name__
		register_component(name, cmp)
		return cmp
	
	return _cmp


def AutoComponent(name=None, aliases=None):
	'''
	Instead of directly passing the config to an AutoComponent, the necessary args are auto filled and passed in.
	This means AutoComponents are somewhat limited in that they cannot modify the config object and they cannot be
	modified with AutoModifiers.

	Note: AutoComponents are usually components that are created with functions (rather than classes) since they can't
	be automodified. When registering classes as components, you should probably use `Component` instead, and pull
	from the config directly.

	:param name: name to use when registering the auto component
	:param aliases: optional aliases for arguments used when autofilling (should be a dict[name,list[aliases]])
	:return: decorator function
	'''
	
	def _auto_cmp(cmp):
		nonlocal name, aliases
		
		if type(cmp) == type:  # to allow AutoModifiers
			
			cls = type('Auto_{}'.format(cmp.__name__), (cmp,), {})
			
			def cmp_init(self, info):
				args, kwargs = autofill_args(cmp, info, aliases=aliases, run=False)
				super(cls, self).__init__(*args, **kwargs)
			
			cls.__init__ = cmp_init
			
			_create = cls
		
		else:
			def _create(config):
				nonlocal cmp, aliases
				return autofill_args(cmp, config, aliases)
		
		Component(name)(_create)
		
		return cmp
	
	return _auto_cmp

# endregion

# region Modifier

_appendable_keys = {'_mod'} # mods can be appended when updating (instead of overwriting)
# def register_appendable_key(key): # At that point, they might as well manipulate _appendable_keys directly
# 	_appendable_keys.add(key)

class _Modifier_Registry(Entry_Registry, components=['fn', 'expects_config', 'project']):
	pass
_mod_registry = _Modifier_Registry()

def register_modifier(name, mod_fn, expects_config=False):
	'''
	Takes as input the "create_fn" of some component and a Config object.

	NOTE: modifier names may not start with "+", as that is used to signify that the mod should be
	appended when merging configs

	:param name: str (should be unique to this component, may not start with "+")
	:param mod_fn: callable accepting one arg (the "create_fn" of a registered component) (these should usually be classes)
	'''
	# assert name not in _reserved_names, '{} is reserved'.format(name)
	# prt.debug(f'Registering modifier {name}')
	# if name in _mod_registry:
	# 	prt.warning(f'A modifier with name {name} has already been registered, now overwriting')
	
	project = get_current_project()
	_mod_registry.new(name, fn=mod_fn, expects_config=expects_config, project=project)
	
	if project is not None:
		project.new_modifier(name)
	
def Modifier(name=None, expects_config=False):
	'''
	Decorator to register a modifier

	NOTE: a :class:`Modifier` is usually not a type/class, but rather a function
	(except :class:`AutoModifiers`, see below)

	:param name: if not provided, will use the __name__ attribute.
	:param expects_config: True iff this modifier expects to be given the config as second arg
	:return: decorator function
	'''
	def _mod(mod):
		nonlocal name, expects_config
		if name is None:
			name = mod.__name__
		register_modifier(name, mod, expects_config=expects_config)
		return mod
	return _mod

def AutoModifier(name=None):
	'''
	Can be used to automatically register modifiers that combine types

	To keep component creation as clean as possible, modifier types should allow arguments to their __init__
	(other than the Config object) and only call pull on arguments not provided, that way child classes of
	the modifier types can specify defaults for the modifications without calling pull() multiple times
	on the same arg.

	Note: in a way, this converts components to modifiers (but think before using). This turns the modified
	component into a child class of this modifier and its previous type.

	In short, Modifiers are used for wrapping of components, AutoModifiers are used for subclassing components

	:param name: if not provided, will use the __name__ attribute.
	:return: decorator to decorate a class
	'''

	def _auto_mod(mod_type):
		nonlocal name
		def _the_mod(cmpn_type):
			# awesome python feature -> dynamic type declaration!
			return type('{}_{}'.format(mod_type.__name__, cmpn_type.__name__), (mod_type, cmpn_type), {})
		Modifier(name=mod_type.__name__ if name is None else name)(_the_mod)
		return mod_type

	return _auto_mod

def _make_post_mod(mod):
	def _make_cmpn(cmpn):
		def _make_mod(info):
			return mod(cmpn(info), info)
		return _make_mod
	return _make_cmpn

def Modification(name=None):
	'''
	A kind of Modifier that modifies the component after it is created,
	and then returns the modified component
	
	expects a callable with input (component, config)
	
	Modifications should almost always be applied after all other modifiers,
	so they should appear at the end of _mod list
	
	:param name: name to register
	:return: a decorator expecting the modification function
	'''
	
	def _reg_modification(mod):
		nonlocal name
		Modifier(name)(_make_post_mod(mod))
		return mod
	
	return _reg_modification

# endregion


def create_component(info):
	'''
	Creates the component specified in info (checks component registry using info.pull('_type'),
	and modifier registry for info.pull('_mod'))

	_mod can be a list, in which case they will be applied in the given order, eg:

	let mods = [A, B, C]

	component <- C(B(A(component)))

	_mod can also be a dict, in which case the keys should be the mod names and the values the order (low to high).
	So for the same behavior as above, a _mod could also be {A:0, B:1, C:2}

	NOTE: generally, modifiers should be ordered from more specific to more general

	:param info: should be a config object with attribute "_type" (and optionally "_mod")
	:return: component created using the provided config (``info``)
	'''

	name = info.pull('_type', silent=True)

	assert name in _component_registry, 'Unknown component type (have you registered it?): {}'.format(name)

	component = _component_registry[name].create_fn

	mod_names = info.pull('_mod', None, silent=True)
	if mod_names is not None:

		if isinstance(mod_names, dict):
			mod_names = sorted(mod_names.keys(), key=lambda k:mod_names[k])

		if not isinstance(mod_names, (list, tuple)):
			mod_names = mod_names,

		for mod_name in mod_names: # WARNING: apply modifications in reverse order
			if mod_name[0] == '+': # filter out initial + if found
				mod_name = mod_name[1:]
			mod = _mod_registry[mod_name]
			component = mod.fn(component, info) if mod.expects_config else mod.fn(component)

	return component(info)



def view_script_registry():
	'''Returns a copy of the full script registry'''
	return _script_registry.copy()

def view_component_registry():
	'''Returns a copy of the full component registry'''
	return _component_registry.copy()

def view_modifier_registry():
	'''Returns a copy of the full modifier registry'''
	return _mod_registry.copy()




import sys, os
import humpack as hp
import io, yaml, json
from collections import defaultdict, OrderedDict
from c3linearize import linearize

from omnibelt import load_yaml, get_printer

from .util import primitives
from .errors import YamlifyError, MissingConfigError, UnknownActionError, InvalidKeyError
from .external import find_config_path
from .registry import create_component, _appendable_keys, Component

nones = {'None', 'none', '_none', '_None', 'null', 'nil', }

prt = get_printer(__name__)

def load_config_from_path(path, process=True):
	'''
	Load the yaml file and transform data to a config object
	
	Generally, ``get_config`` should be used instead of this method
	
	:param path: must be the full path to a yaml file
	:param process: if False, the loaded yaml data is passed without converting to a config object
	:return: loaded data from path (usually as a config object)
	'''
	# path = find_config_path(path)
	data = load_yaml(path)
	
	if data is None:
		data = {}
	
	if process:
		return configurize(data)
	return data

def process_single_config(data, process=True, parents=None):  # data can either be an existing config or a path to a config
	'''
	This loads the data (if a path or name is provided) and then checks for parents and loads those as well
	
	Generally, ``get_config`` should be used instead of this method
	
	:param data: config name or path or raw data (dict/list) or config object
	:param process: configurize loaded data
	:param parents: if None, no parents are loaded, otherwise it must be a dict where the keys are the absolute paths to the config (yaml) file and values are the loaded data
	:return: loaded data (as a config object or raw)
	'''
	
	if isinstance(data, str):
		data = find_config_path(data)
		data = load_config_from_path(data, process=process)
	
	if parents is not None and 'parents' in data:
		todo = []
		for parent in data['parents']:  # prep new parents
			# ppath = _config_registry[parent] if parent in _config_registry else parent
			ppath = find_config_path(parent)
			if ppath not in parents:
				todo.append(ppath)
				parents[ppath] = None
		for ppath in todo:  # load parents
			parents[ppath] = process_single_config(ppath, parents=parents)
	
	return data

def merge_configs(configs, parent_defaults=True):
	'''
	configs should be ordered from oldest to newest (ie. parents first, children last)
	
	This is an internal method used by ``get_config()`` and should generally not be called manually.
	'''
	
	if not len(configs):
		return ConfigDict()
	
	child = configs.pop()
	merged = merge_configs(configs, parent_defaults=parent_defaults)
	
	# load = child.load if 'load' in child else None
	merged.update(child)
	
	return merged


def get_config(*contents, **manual):  # Top level function
	'''
	Top level function for users. This is the best way to load/create a config object.

	All parent config (registered names or paths) that should be loaded
	must precede any manual entries, and will be loaded in reverse order (like python class inheritance).
	
	If the key ``_history_key`` is specified and not :code:`None`, a flattened list of all parents of
	this config is pushed to the given key.
	
	:param contents: registered configs or paths or manual entries (like in terminal)
	:param manual: specify parameters manually as key value pairs
	:return: config object
	'''
	root = ConfigDict()
	if len(contents) + len(manual) == 0:
		return root

	reg = []
	terms = {**manual}
	allow_reg = True
	waiting_key = None
	
	for term in contents:
		
		if term.startswith('--'):
			allow_reg = False
			if waiting_key is not None:
				terms[waiting_key] = True
			waiting_key = term[2:]
		
		elif waiting_key is not None:
			terms[waiting_key] = process_raw_argv(term)
			waiting_key = None
			
		elif allow_reg:
			reg.append(term)
		
		else:
			raise Exception(f'Parsing error: {term} in {contents}')
		
	if waiting_key is not None:
		terms[waiting_key] = True
		
	root.update(configurize(terms))
	
	if len(reg) == 0:
		return root

	root['parents'] = ConfigList(data=reg + (list(root['parents']) if 'parents' in root else []))

	parents = {}
	
	root = process_single_config(root, parents=parents)

	pnames = []
	if len(parents):  # topo sort parents
		
		# TODO: maybe clean up?
		
		root_id = ' root'
		src = defaultdict(list)
		
		names = {find_config_path(p): p for p in root['parents']} if 'parents' in root else {}
		src[root_id] = list(names.keys())
		
		for n, data in parents.items():
			connections = {find_config_path(p): p for p in data['parents']} if 'parents' in data else {}
			names.update(connections)
			src[n] = list(connections.keys())
		
		order = linearize(src, heads=[root_id], order=True)[root_id]
		
		pnames = [names[p] for p in order[1:]]
		order = [root] + [parents[p] for p in order[1:]]
		
		# for analysis, record the history of all loaded parents
		order = list(reversed(order))
	
	else:  # TODO: clean up
		order = [root]
	
	root = merge_configs(order,)
	
	include_history = root.pull('_history_key', None, silent=True)
	if include_history is not None:
		root.push(include_history, pnames, silent=True)
	
	root.push('parents', '_x_', silent=True)
	
	return root




_printing_instance = None
class Config_Printing:
	'''
	Internal class to manage the printing pulls/pushes of the config object. (eg. indent/line styles)
	'''
	def __new__(cls, *args, **kwargs): # singleton
		global _printing_instance
		if _printing_instance is None:
			_printing_instance = super().__new__(cls)
			
			_printing_instance.level = 0
			_printing_instance.is_new_line = True
			_printing_instance.unit = ' > '
			_printing_instance.style = '| '
			_printing_instance.silent = False
			
		return _printing_instance
	def __repr__(self):
		return f'ConfigPrinting[{hex(id(self))}]'
	def __str__(self):
		return f'ConfigPrinting'
	
	def inc_indent(self):
		self.level += 1
	def dec_indent(self):
		self.level = max(0, self.level-1)
	
	def process_addr(self, *terms):

		# return '.'.join(terms)
		addr = []
		
		skip = False
		for term in terms[::-1]:
			if term == '':
				skip = True
				addr.append('')
			elif term is None:
				pass
			elif not skip:
				addr.append(term)
			else:
				skip = False
		
		if not len(addr):
			return '.'
		
		addr = '.'.join(addr[::-1])
		return addr
	
	def log_record(self, raw, end='\n',
	               silent=False):
		indent = self.level * self.unit
		style = self.style
		prefix = style + indent
		
		msg = raw.replace('\n', '\n' + prefix)
		if not self.is_new_line:
			prefix = ''
		msg = f'{prefix}{msg}{end}'
		base = self.silent
		if not (silent or base):
			print(msg, end='')
			self.is_new_line = (len(msg) == 0 and self.is_new_line) or msg[-1] == '\n'
		return msg
		

class ConfigType(hp.Transactionable):
	'''
	The abstract super class of config objects.
	
	The most important methods:
	
		- ``push()`` - set a parameter in the config
		- ``pull()`` - get a parameter in the config
		- ``sub()`` - get a sub branch of this config
		- ``seq()`` - iterate through all contents
		- ``update()`` - update a config with a different config
		- ``export()`` - save the config object as a yaml file
	
	Another important property of the config object is that it acts as a tree where each node can hold parameters.
	If a parameter cannot be found at one node, it will search up the tree for the parameter.
	
	Config objects also allow for "deep" gets/sets, which means you can get and set parameters not just in
	the current node, but any number of nodes deeper in the tree by passing a list/tuple of keys
	or keys separated by "." (hereafter called the "address" of the parameter).
	
	Note: that all parameters must not contain "." and should generally be valid python identifiers
	(strings with no white space that don't start with a number).
	
	'''

	def __init__(self, parent=None, silent=False, printer=None, prefix=None, safe_mode=False,
	             data=None):
		'''
		Generally it should not be necessary to create a ConfigType manually, instead use ``get_config()``.
		
		:param parent: parent config used for defaults
		:param silent: don't print ``pull``s and ``push``es
		:param printer: printer object to handle printing messages
		:param prefix: initial prefix used for printing
		:param safe_mode: don't save created component instances, unless during a transaction
		:param data: raw parameters to immediately add to this config
		'''
		
		if printer is None:
			printer = Config_Printing()
		self.__dict__['_printer'] = printer
		
		if prefix is None:
			prefix = []
		self.__dict__['_prefix'] = prefix
		self.__dict__['_hidden_prefix'] = prefix.copy()
		self.__dict__['_safe_mode'] = safe_mode
		
		self.set_parent(parent)
		# self._set_silent(silent)

		super().__init__()
		
		if data is not None:
			self.update(data)
		
	def sub(self, item):
		'''
		Used to get a subbranch of the overall config
		:param item: address of the branch to return
		:return: config object at the address
		'''
		
		val = self.get_nodefault(item)
		
		if isinstance(item, (list, tuple)):
			item = '.'.join(item)
		
		if isinstance(val, ConfigType):
			val._set_prefix(self.get_prefix() + [item])
			val._store_prefix()
		
		return val

	def update(self, other):
		'''
		Used to merge two config nodes (and their children) together
		
		This method must be implemented by child classes depending on how the contents of the node is stored
		
		:param other: config node to overwrite ``self`` with
		:return: None
		'''
		try:
			return super().update(other)
		except AttributeError:
			raise NotImplementedError

	def _record_action(self, action, suffix=None, val=None, silent=False, obj=None,
	                   defaulted=False, pushed=False, _entry=False):
		'''
		Internal function to manage printing out messages after various actions have been taken with this config.
		
		:param action: name of the action (see code for examples)
		:param suffix: suffix of the address (aka. last item in the address)
		:param val: contents at that address
		:param silent: suppress printing a message for this action
		:param obj: object created or used in this action (eg. a newly created component)
		:param defaulted: is a default value being used
		:param pushed: has this value just been pushed
		:param _entry: internal flag used for printing messages
		:return: formatted message corresponding to this action
		'''
		printer = self._get_printer()

		if action == 'defaulted':
			return ''
		
		name = printer.process_addr(*self.get_prefix(), suffix)
		
		if pushed:
			name = '[Pushed] ' + name

		if 'alias' in action:
			return printer.log_record(f'{name} --> ', end='', silent=silent)
		
		origins = ' (by default)' if defaulted else ''
		
		if action == 'removing':
			obj_type = action.split('-')[-1]
			return printer.log_record(f'REMOVING {name}', silent=silent)
		
		if action in {'creating', 'reusing'}:
			
			assert val is not None, 'no info provided'
			
			cmpn_type = val.pull('_type', silent=True)
			mods = val.pull('_mod', None, silent=True)
			
			mod_info = ''
			if mods is not None:
				if not isinstance(mods, (list, tuple)):
					mods = mods,
				
				mod_info = ' (mods=[{}])'.format(', '.join(m for m in mods)) if len(mods) > 1 \
					else f' (mod={mods[0]})'
			
			end = ''
			if action == 'reusing':
				assert obj is not None, 'no object provided'
				end = f': id={hex(id(obj))}'
			
			# head = '' if suffix is None else f' {name}'
			head = f' {name}'
			out = printer.log_record(f'{action.upper()}{head} '
			                                               f'(type={cmpn_type}){mod_info}{origins}{end}',
			                       silent=silent)
			if action == 'creating':
				printer.inc_indent()
			return out
		
		if action in {'iter-dict', 'iter-list'}:
			obj_type = action.split('-')[-1]
			
			head = '' if suffix is None else f'{name} [{obj_type} with {len(val)} item/s]'
			return printer.log_record(f'ITERATOR {head}{origins}', silent=silent)
		
		if action in {'pull-dict', 'pull-list'}:
			
			assert val is not None, 'no obj provided'
			
			obj_type = action.split('-')[-1]
			
			out = printer.log_record(f'{name} [{obj_type} with {len(val)} item/s]', silent=silent)
			printer.inc_indent()
			return out
		
		if action in {'created', 'pulled-dict', 'pulled-list'}:
			assert obj is not None, 'no object provided'
			printer.dec_indent()
			return ''
			return printer.log_record(f'=> id={hex(id(obj))}', silent=silent)

		pval = None if val is None else repr(val)
		
		if action == 'entry': # when pulling dict/list
			assert suffix is not None, 'no suffix provided'
			return printer.log_record(f'({suffix}): ', end='',
			                          silent=(silent or self._get_silent()))
		
		if action == 'pulled':
			head = '' if suffix is None else f'{name}: '
			if _entry:
				head = ''
			return printer.log_record(f'{head}{pval}{origins}', silent=silent)
		
		raise UnknownActionError(action)
		
	def pull(self, item, *defaults, silent=False, ref=False, no_parent=False, as_iter=False):
		'''
		Top-level function to get parameters from the config object (including automatically creating components)

		:param item: address of the parameter to get
		:param defaults: default values to use if ``item`` is not found
		:param silent: suppress printing message that this parameter was pulled
		:param ref: if the parameter is a component that has already been created, get a reference to the created component instead of creating a new instance
		:param no_parent: don't default to a parent node if the ``item`` is not found here
		:param as_iter: return an iterator over the selected value (only works if the value is a dict/list)
		:return: processed value of the parameter (or default if ``item`` is not found, or raises a ``MissingConfigError`` if not found)
		'''
		self._reset_prefix()
		return self._pull(item, *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter, _origin=self)

	def pull_self(self, name=None, silent=False, as_iter=False):
		'''
		Process self as a value being pulled.
		
		:param name: Name given to self for printed message
		:param silent: suppress printing message
		:param as_iter: Return self as an iterator (has same effect as calling ``seq()``)
		:return: the processed value of self
		'''
		self._reset_prefix()
		return self._process_val(name, self, silent=silent, reuse=False, is_self=True, as_iter=as_iter, _origin=self)

	def _pull(self, item, *defaults, silent=False, ref=False, no_parent=False, as_iter=False,
	          _defaulted=False, _origin=None):
		'''
		Internal pull method, should generally not be called manually (unless you know what you're doing)
		
		:param item: remaining address to find
		:param defaults: any default values that can be used if address is not found
		:param silent: suppress messages
		:param ref: return an instance of the value instead of creating a new instance, if one exists
		:param no_parent: do not check for the parameter in the parent
		:param as_iter: return the value as an iterator (only for dicts/lists)
		:param _defaulted: flag that this value was once a provided default (used for printing)
		:param _origin: reference to the original config node that was pulled (some pulls require returing to origin)
		:return: processed value found at ``item`` or a processed default value
		'''
		
		# TODO: change defaults to be a keyword argument providing *1* default, and have item*s* instead,
		#  which are the keys to be checked in order

		if '.' in item:
			item = item.split('.')

		line = []
		if isinstance(item, (list, tuple)):
			line = item[1:]
			item = item[0]

		defaulted = item not in self
		byparent = not self.contains_nodefault(item)
		if no_parent and byparent:
			defaulted = True
		if defaulted:
			if len(defaults) == 0:
				raise MissingConfigError(item)
			val, *defaults = defaults
		else:
			val = self[item]

		if len(line) and not isinstance(val, ConfigType):
			defaulted = True
			if len(defaults) == 0:
				raise MissingConfigError(item)
			val, *defaults = defaults

		if defaulted: # try again with new value

			_origin._reset_prefix()
			
			val = _origin._process_val(item, val, *defaults, silent=silent, defaulted=defaulted or _defaulted,
			                        as_iter=as_iter, reuse=ref, _origin=_origin)

		elif len(line): # child pull
			if not isinstance(val, ConfigType):
				prt.warning(f'Pulling through a non-config object: {val}')
			
			val._set_prefix(self.get_prefix() + [item])
			out = val._pull(line, *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter,
			               _defaulted=_defaulted, _origin=_origin)
			
			val = out
			
		elif byparent: # parent pull
			parent = self.get_parent()

			parent._set_prefix(self.get_prefix() + [''])
			val = parent._pull((item, *line), *defaults, silent=silent, ref=ref, no_parent=no_parent, as_iter=as_iter,
			                   _origin=_origin)
			
		else: # process val from me/defaults
			val = self._process_val(item, val, *defaults, silent=silent, defaulted=defaulted or _defaulted,
			                        as_iter=as_iter, reuse=ref, _origin=_origin)
			
			if type(val) in {list, set}: # TODO: a little heavy handed
				val = tuple(val)

		return val

	def _process_val(self, item, val, *defaults, silent=False,
	                 reuse=False, is_self=False, as_iter=False, _origin=None, **record_flags):
		'''
		This is used by ``pull()`` to process the recovered value and print the correct message if ``not silent``
		
		:param item: remaining address where this value was found
		:param val: value that was found given the original address
		:param defaults: any additional defaults to use if processing fails with ``val``
		:param silent: suppress messages
		:param reuse: if an instance is found (under ``__obj``) then that should be returned instead of creating a new one
		:param is_self: this config object should be returned after processing
		:param as_iter: return an iterator of ``val`` (only works if ``val`` is a list/dict)
		:param _origin: original config node where the pull or push request was intially called
		:param record_flags: additional flags used for printing.
		:return: processed value
		'''
		
		if as_iter and isinstance(val, ConfigType):
			
			obj_type = 'list' if isinstance(val, list) else 'dict'
			
			self._record_action(f'iter-{obj_type}', suffix=item, val=val, silent=silent, **record_flags)
			
			itr = Config_Iter(val, val)
			
			return itr
		
		if isinstance(val, dict):
			if '_type' in val:
				
				# val.push('__origin_key', item, silent=True)
				
				if reuse and '__obj' in val:
					# print('WARNING: would usually reuse {} now, but instead creating a new one!!')
					cmpn = val['__obj']
					self._record_action('reusing', suffix=item, val=val, obj=cmpn, silent=silent, **record_flags)
				else:

					self._record_action('creating', suffix=item, val=val, silent=silent, **record_flags)
					
					hidden = val._get_hidden_prefix()
					val.get_prefix().clear()
					val._store_prefix()
					
					past = self._get_silent()
					with self.silenced(silent or past):
						cmpn = create_component(val)
					
					if not self.in_safe_mode() or self.in_transaction():
						if item is not None and len(item) and not is_self:
							self[item]['__obj'] = cmpn
						else:
							self['__obj'] = cmpn
					
					val._set_prefix(hidden)
					val._store_prefix()
					
					self._record_action('created', suffix=item, val=val, obj=cmpn, silent=silent, **record_flags)
					
				val = cmpn

			else:
				
				self._record_action('pull-dict', suffix=item, val=val, silent=silent, **record_flags)
				terms = {}
				for k, v in val.items():  # WARNING: pulls all entries in dict
					self._record_action('entry', silent=silent, suffix=k)
					terms[k] = val._process_val(k, v, reuse=reuse, silent=silent, _origin=_origin, _entry=True)
				self._record_action('pulled-dict', suffix=item, val=val, obj=terms, silent=silent, **record_flags)
				val = terms


		elif isinstance(val, list):
			
			self._record_action('pull-list', suffix=item, val=val, silent=silent, **record_flags)
			terms = []
			for i, v in enumerate(val):  # WARNING: pulls all entries in list
				self._record_action('entry', silent=silent, suffix=str(i))
				terms.append(val._process_val(str(i), v, reuse=reuse, silent=silent, _origin=_origin, _entry=True))
			self._record_action('pulled-list', suffix=item, val=val, obj=terms, silent=silent, **record_flags)
			val = terms

		elif isinstance(val, str) and val.startswith('<>'):  # local alias (looks for alias locally)
			alias = val[2:]
			
			self._record_action('local-alias', suffix=item, val=alias, silent=silent, **record_flags)
			
			val = self._pull(alias, *defaults, silent=silent, _origin=_origin)
			
		elif isinstance(val, str) and val.startswith('<o>'): # origin alias (returns to origin to find alias)
			alias = val[3:]
			
			self._record_action('origin-alias', suffix=item, val=alias, silent=silent, **record_flags)
			
			_origin._reset_prefix()
			val = _origin._pull(alias, *defaults, silent=silent, _origin=_origin)
			
		else:
			self._record_action('pulled', suffix=item, val=val, silent=silent, **record_flags)

		return val
	
	def push(self, key, val, *_skip, silent=False, overwrite=True, no_parent=True, force_root=False):
		'''
		Set ``key`` with ``val`` in the config object, but pulls ``key`` first so that `val` is only set
		if it is not found or ``overwrite`` is set to ``True``. It will return the current value of ``key`` after
		possibly setting with ``val``.

		:param key: key to check/set (can be list or '.' separated string)
		:param val: data to possibly write into the config object
		:param _skip: soak up all other positional arguments to make sure the remaining are keyword only
		:param silent: Do not print messages
		:param overwrite: If key is already set, overwrite with (configurized) 'val'
		:param no_parent: Do not check parent object if not found in self
		:param force_root: Push key to the root config object
		:return: current val of key (updated if written)
		'''
		self._reset_prefix()
		return self._push(key, val, silent=silent, overwrite=overwrite, no_parent=no_parent, force_root=force_root)
	
	def _push(self, key, val, silent=False, overwrite=True, no_parent=True, force_root=False, _origin=None):
		'''
		Set ``key`` with ``val`` in the config object, but pulls ``key`` first so that `val` is only set
		if it is not found or ``overwrite`` is set to ``True``. It will return the current value of ``key`` after
		possibly setting with ``val``.
		
		:param key: key to check/set (can be list or '.' separated string)
		:param val: data to possibly write into the config object
		:param _skip: soak up all other positional arguments to make sure the remaining are keyword only
		:param silent: Do not print messages
		:param overwrite: If key is already set, overwrite with (configurized) 'val'
		:param no_parent: Do not check parent object if not found in self
		:param force_root: Push key to the root config object
		:return: current val of key (updated if written)
		'''
		
		if '.' in key:
			key = key.split('.')

		line = []
		if isinstance(key, (list, tuple)):
			line = key[1:]
			key = key[0]
		
		exists = self.contains_nodefault(key)
		
		parent = self.get_parent()
		
		if parent is not None and force_root:
			return self.get_root().push((key, *line), val, silent=silent, overwrite=overwrite,
			                   no_parent=no_parent)
		elif no_parent:
			parent = None
		
		if exists and len(line): # push me
			child = self.get_nodefault(key)
			
			child._set_prefix(self.get_prefix() + [key])
			
			out = child._push(line, val, silent=silent, overwrite=overwrite, no_parent=True)
			
			return out
		elif parent is not None and key in parent: # push parent

			parent._set_prefix(self.get_prefix() + [''])
			
			out = parent._push((key, *line), val, silent=silent, overwrite=overwrite, no_parent=no_parent)
			
			return out
			
		elif len(line): # push child
			
			child = self.get_nodefault(key)
			
			child._set_prefix(self.get_prefix() + [key])
			
			out = child._push(line, val, silent=silent, overwrite=overwrite, no_parent=True)
		
			return out
		
		if exists and not overwrite:
			return self._pull(key, silent=True)
		
		if exists and isinstance(val, str) and val == '_x_':
			self._record_action('removing', suffix=key, val=val, silent=silent)
			del self[key]
			return
		
		val = configurize(val)
		# val = self.__setitem__(key, val)
		
		self[key] = val
		# val = self[key]
		
		val = self._process_val(key, val, silent=silent, pushed=True)
		
		return val

	def export(self, path=None):
		'''
		Convert all data to raw data (using dict/list) and save as yaml file to ``path`` if provided.
		Also returns yamlified data.
		
		:param path: path to save data in this config (data is not saved to disk if not provided)
		:return: raw "yamlified" data
		'''

		data = yamlify(self)

		if path is not None:
			if os.path.isdir(path):
				path = os.path.join(path, 'config.yaml')
			with open(path, 'w') as f:
				yaml.dump(data, f)
			return path

		return data
	
	def seq(self):
		'''
		Returns an iterator over the contents of this config object where elements are lazily
		processed during iteration (see ``Config_Iter`` for details).
		
		:return: iterator over all arguments in self
		'''
		return Config_Iter(self, self)
	
	# region Silencing
	
	def _set_silent(self, silent=True):
		self.__dict__['_printer'].silent = silent
	
	# self._silent_config_flag = silent
	
	def _get_silent(self):
		return self.__dict__['_printer'].silent
	
	class _Silent_Config:
		'''Internal context manager to silence a config object'''
		
		def __init__(self, config, setting):
			self.config = config
			self.setting = setting
			self.prev = config._get_silent()
		
		def __enter__(self):
			self.config._set_silent(self.setting)
			return self.config
		
		def __exit__(self, exc_type, exc_val, exc_tb):
			self.config._set_silent(self.prev)
	
	def silenced(self, setting=True):
		return ConfigType._Silent_Config(self, setting=setting)
	
	# endregion
	
	# region Parents
	
	def is_root(self):  # TODO: move to tree space
		'''Check if this config object has a parent for defaults'''
		return self.get_parent() is None
	
	def set_parent(self, parent):
		self.__dict__['_parent'] = parent
	
	def get_parent(self):
		'''Get parent (returns None if this is the root)'''
		return self.__dict__['_parent']
	
	def get_root(self):
		'''Gets the root config object (returns ``self`` if ``self`` is the root)'''
		parent = self.get_parent()
		if parent is None:
			return self
		return parent.get_root()
	
	# endregion
	
	# region Addressing
	
	def _get_printer(self):
		return self.__dict__['_printer']
	
	def get_prefix(self):
		return self.__dict__['_prefix']
	
	def _set_prefix(self, prefix):
		self.__dict__['_prefix'] = prefix
	
	def _reset_prefix(self):
		self._set_prefix(self._get_hidden_prefix())
	
	def _get_hidden_prefix(self):
		return self.__dict__['_hidden_prefix']
	
	def _store_prefix(self):
		self.__dict__['_hidden_prefix'] = self.get_prefix().copy()
	
	# endregion
	
	# region Misc
	
	def set_safe_mode(self, safe_mode):
		if self.is_root():
			self.__dict__['_safe_mode'] = safe_mode
		else:
			self.get_root().set_safe_mode(safe_mode)
	
	def in_safe_mode(self):
		if self.is_root():
			return self.__dict__['_safe_mode']
		return self.get_root().in_safe_mode()
	
	@staticmethod
	def parse_argv(arg):
		try:
			return yaml.safe_load(io.StringIO(arg))
		except:
			pass
		return arg
	
	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"

		Must be implemented by the child class

		:return: None
		'''
		raise NotImplementedError
	
	def __repr__(self):
		return f'{type(self).__name__}[id={hex(id(self))}]'
	
	def __str__(self):
		return f'<{type(self).__name__}>'
	
	# endregion
	
	# region Get/Set/Contains
	
	def __setitem__(self, key, value):
		if isinstance(key, str) and '.' in key:
			key = key.split('.')
		
		if isinstance(key, (list, tuple)):
			if len(key) == 1:
				return self.__setitem__(key[0], value)
			child = self.__getitem__(key[0])
			assert isinstance(child, ConfigType)
			return child.__setitem__(key[1:], value)
		
		value = configurize(value)
		
		if isinstance(value, ConfigType):
			value.set_parent(self)
		return super().__setitem__(key, value)
	
	def __getitem__(self, item):
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (list, tuple)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.__getitem__(item[0])[item[1:]]
		
		parent = self.get_parent()
		
		if not self.contains_nodefault(item) \
				and parent is not None \
				and item[0] != '_':
			return parent[item]
		
		return self._single_get(item)
	
	def get_nodefault(self, item):
		'''Get ``item`` without defaulting up the tree if not found.'''
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (list, tuple)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.get_nodefault(item[0])[item[1:]]
		
		return self._single_get(item)
	
	def _single_get(self, item):
		try:
			val = super().__getitem__(item)
		except KeyError:
			return self._missing_key(item)
		
		return val
	
	def _missing_key(self, key):
		obj = self.__class__(parent=self)
		self.__setitem__(key, obj)
		return obj
	
	def __contains__(self, item):
		'''Check if ``item`` is in this config, item can be "deep" (multiple steps'''
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (tuple, list)):
			if len(item) == 1:
				item = item[0]
			else:
				return item[0] in self and item[1:] in self[item[0]]
		
		parent = self.get_parent()
		
		return self.contains_nodefault(item) \
		       or (not super().__contains__(item)
		           and parent is not None
		           and item[0] != '_'
		           and item in parent)
	
	def contains_nodefault(self, item):
		'''Check if ``item`` is contained in this config object without defaulting up the tree if ``item`` is not found'''
		
		if isinstance(item, str) and '.' in item:
			item = item.split('.')
		
		if isinstance(item, (tuple, list)):
			if len(item) == 1:
				item = item[0]
			else:
				return self.contains_nodefault(item[0]) and self[item[0]].contains_nodefault(item[1:])
		
		if super().__contains__(item):
			return self.get_nodefault(item) is not '__x__'
		return False

	# endregion


class ConfigDict(ConfigType, hp.tdict):
	'''
	Dict like node in the config.
	
	Keys should all be valid python attributes (strings with no whitespace, and not starting with a number).

	NOTE: avoid setting keys that start with more than one underscore (especially '__obj')
	(unless you really know what you are doing)
	'''
	
	def update(self, other):
		'''
		Merge self with another dict-like object
		
		:param other: must be dict like
		:return: None
		'''
		assert isinstance(other, dict), f'invalid: {type(other)}'
		
		for k, v in other.items():
			isconfig = isinstance(v, ConfigType)
			exists = self.contains_nodefault(k)
			if exists and '_x_' == v:  # reserved for deleting settings in parents
				del self[k]
			
			elif exists and isconfig and \
					(isinstance(v, self[k].__class__)
					 or isinstance(self[k], v.__class__)):
				self[k].update(v)
				
			else:
				self[k] = v
	
	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"
		
		:return: None
		'''
		bad = []
		for k, v in self.items():
			if k.startswith('__'):
				bad.append(k)
			elif isinstance(v, ConfigType):
				v.purge_volatile()
		
		for k in bad:
			del self[k]
	
	def __repr__(self):
		info = '{{' + ', '.join(f'{k}' for k in self) + '}}'
		return f'[{hex(id(self))}]{info}'
	
	def __str__(self):
		return '{{' + ', '.join(f'{k}' for k in self) + '}}'




class ConfigList(ConfigType, hp.tlist):
	'''
	List like node in the config.
	'''

	def __init__(self, *args, empty_fill_value=None, **kwargs):
		super().__init__(*args, **kwargs)
		self._empty_fill_value = empty_fill_value

	def purge_volatile(self):
		'''
		Recursively remove any items where the key starts with "__"
		
		:return: None
		'''
		for x in self:
			if isinstance(x, ConfigType):
				x.purge_volatile()
		
	def _str_to_int(self, item):
		'''Convert the input items to indices of the list'''
		if isinstance(item, int):
			return item

		try:
			return int(item)
		except (TypeError, ValueError):
			pass

		raise InvalidKeyError(f'failed to convert {item} to an index')
		
	def update(self, other):
		'''Overwrite ``self`` with the provided list ``other``'''
		for i, x in enumerate(other):
			isconfig = isinstance(x, ConfigType)
			if len(self) == i:
				self.append(x)
			elif isconfig and (isinstance(x, self[i].__class__) or isinstance(self[i], x.__class__)):
				self[i].update(x)
			else:
				self[i] = x
	
	def _single_get(self, item):
		try:
			idx = self._str_to_int(item)
		except InvalidKeyError:
			idx = None
		
		return super()._single_get(item if idx is None else idx)

	def push(self, first, *rest, overwrite=True, **kwargs):
		'''
		When pushing to a list, if you don't provide an index, the value is automatically pushed to the end of the list
		
		:param first: if no additional args are provided in `rest`, then this is used as the value and the key is the end of the list, otherwise this is used as key and the first element in `rest` is the value
		:param rest: optional second argument to specify the key, rather than defaulting to the end of the list
		:param kwargs: same keyword args as for the ConfigDict
		:return: same as for ConfigDict
		'''

		if len(rest):
			val, *rest = rest
			return super().push(first, val, *rest, overwrite=overwrite, **kwargs)
		
		val = first
		key = len(self)
		self.append(None)
		return super().push(key, val, *rest, overwrite=True, **kwargs) # note that *rest will have no effect
		
	def __setitem__(self, key, value):

		if key == '_': # append to end of the list
			return self.__setitem__(len(self), value)

		idx = self._str_to_int(key)
		
		if idx >= len(self):
			self.extend([self._empty_fill_value]*(idx-len(self)+1))
		return super().__setitem__(idx, value)
		
		
	def contains_nodefault(self, item):
		
		try:
			idx = self._str_to_int(item)
		except InvalidKeyError:
			return isinstance(item, slice)
		
		N = len(self)
		return -N <= idx < N
		
	def append(self, item):
		super().append(item)
		if isinstance(item, ConfigType):
			item.set_parent(self)
		
	def extend(self, item):
		super().extend(item)
		
		for x in item:
			if isinstance(x, ConfigType):
				x.set_parent(self)

@Component('iter')
class Config_Iter:
	'''
	Iterate through a list of parameters, processing each item lazily,
	ie. only when it is iterated over (with ``next()``)
	'''

	def __init__(self, origin, elements=None):
		'''
		Can be used as a component or created manually (by providing the ``elements`` argument explicitly)
		
		For dicts, this will behave like ``.items()``, ie. for each entry in the dict it will return
		a tuple of the key and value.
		
		:param origin: config object where the iterator info is
		:param elements: manually provided elements to iterate over (uses contents of "_elements" in ``origin`` if not provided)
		'''
		self._idx = 0
		# self._name = config._ident
		# assert '_elements' in config, 'No elements found'
		if elements is None:
			elements = origin._elements
		self._elms = elements
		
		self._keys = [k for k in self._elms.keys()
		              if k not in {'_elements', '_mod', '_type', '__obj'}
					  and self._elms[k] != '__x__'] \
			if isinstance(self._elms, dict) else None
		self._prefix = origin.get_prefix().copy()

	def __len__(self):
		'''Returns the remaining length of this iterator instance'''
		return len(self._elms if self._keys is None else self._keys) - self._idx

	def _next_idx(self):
		'''Find the next index or key'''
		
		if self._keys is None:
			return str(self._idx)
		
		while self._idx < len(self._elms):
			idx = self._keys[self._idx]
			
			if self._elms.contains_nodefault(idx):
				return idx
			self._idx += 1
		
		raise StopIteration
		
	def view(self):
		'''Returns the next object without processing the item, may throw a StopIteration exception'''
		obj = self._elms[self._next_idx()]
		if isinstance(obj, ConfigType):
			return self._elms.sub(self._next_idx())
		return obj

	def __next__(self):

		if len(self._elms) == self._idx:
			raise StopIteration
		
		idx = self._next_idx()

		self._elms._prefix = self._prefix
		
		obj = self._elms._pull(idx)
		
		self._idx += 1
		if self._keys is None:
			return obj
		return idx, obj

	def __iter__(self):
		return self


def configurize(data):
	'''
	Transform data container to use config objects (ConfigDict/ConfigList)

	:param data: dict/list data
	:return: deep copy of data using ConfigDict/ConfigList
	'''
	if isinstance(data, ConfigType):
		return data
	if isinstance(data, dict):
		return ConfigDict(data={k: configurize(v) for k, v in data.items()})
	if isinstance(data, (list, set)):
		return ConfigList(data=[configurize(x) for x in data])
	if isinstance(data, str) and data in nones:
		return None
	return data


def yamlify(data):  # TODO: allow adding yamlify rules for custom objects
	'''
	Transform data container into regular dicts/lists to export to yaml file

	:param data: Config object
	:return: deep copy of data using dict/list
	'''
	# if data is None:
	# 	return '_None'
	if data is None or isinstance(data, primitives):
		return data
	if isinstance(data, dict):
		return {k: yamlify(v) for k, v in data.items() if not k.startswith('__')}
	if isinstance(data, (list, tuple, set)):
		return [yamlify(x) for x in data]
	
	raise YamlifyError(data)


# _config_type = ConfigType
# _config_dict = ConfigDict
# _config_list = ConfigList

def process_raw_argv(arg):
	'''
	This is the preferred way to parse arguments for the config
	(especially arguments specified through the terminal)
	'''
	return ConfigType.parse_argv(arg)


